/******************************************************************************
*
* project name: TI-MineSweeper
* filename:     timiner.c
* initial date: 04/07/2000
* authors:      thomas.nussbaumer@gmx.net (coding)
*               marcos.lopez@gmx.net (game play/graphics/BETA testing)
*               lionel_debroux@yahoo.fr (speedups/size optimizations)
*
* description:  main file (contents most of the stuff)
*
* Current implementation uses TIGCC 0.95 Beta 12
*
* $Id: timiner.c,v 1.3 2004/07/14 15:08:58 DEBROUX Lionel Exp $
*
******************************************************************************/

//------------------------------------------------------------------------------
//KEYS:
//
// [F1] or [HOME]          ... open a tile
// [F2], [MODE] or [ENTER] ... mark/unmark tile as bomb
// [ESC]                   ... abort game/back to menu
//
//-----------------------------------------------------------------------------

#include "timiner.h"

/*---------------------------------------------------------------------------*/
/*---------------------------------------------------------------------------*/
/* global variables                                                          */
/*---------------------------------------------------------------------------*/
/*---------------------------------------------------------------------------*/
SCR_RECT       fullscreen     = {{0,0,239,127}}; // used for fullscreen clipping
stack_t*       stack          =  0; // stack used for open recursively feature
unsigned char* cursor_buf1    =  0; // background buffer for plane1 of cursor
unsigned char* cursor_buf2    =  0; // background buffer for plane2 of cursor
short          stack_counter  = -1; // counter of entries on the stack
short          nr_columns     =  0; // active board width
short          nr_bombs       =  0; // active number of distributed bombs
short          draw_all       =  1; // indicator if all tiles should be drawn
counter_t      bomb_counter   = {0/*,bc_ypos*/,0,{0,0,0}}; // counter for bombs
counter_t      time_counter   = {0/*,bc_ypos*/,0,{0,0,0}}; // counter for seconds passed
short          mode           = 0;
short          timer_started  = 0;

short tiles_offset_x = 0;
short smiley_x = 0;
short board_offset_x = 0;


/*===========================================================================*/
/* get input key pressed by user                                             */
/*                                                                           */
/* replacement for ngetchx() and/or GKeyIn(), because:                       */
/*                                                                           */
/* (1) ngetchx() won't check the APD timer value                             */
/* (2) GKeyIn() destroys the display due to setting the status display       */
/*                                                                           */
/* additionally this function handles the CALCULATOR OFF combinations        */
/*===========================================================================*/
unsigned short GetUserInput(void) {
    void           *kbq = kbd_queue();
    unsigned short  key;
    short           gray = GrayCheckRunning();

    //-------------------------------------------
    // restart the APD timer now
    //-------------------------------------------
    OSTimerRestart(APD_TIMER);

    while (1) {
        //---------------------------------------
        // wait until we get a key
        // if the APD timer expires -> turn
        // calculator off
        //---------------------------------------
        while (OSdequeue(&key, kbq)) {
            if (!gray) idle();
            else pokeIO(0x600005,0b10111);
            if (OSTimerExpired(APD_TIMER)) {
                off();
                OSTimerRestart(APD_TIMER);
            }

            //-------------------------
            // handling of time counter
            //-------------------------
            if (timer_started) { // only if we have started it by our own
                if (OSTimerExpired(USER_TIMER)) {
                    if (time_counter.value < 999) {
                        time_counter.value++;
                        UpdateCounter(&time_counter);
                    }
                }
            }
        }

        //---------------------------------------
        // handle calculator-off key combinations
        // if the APD timer expires
        // [2nd] + [ON] & [diamond]+[ON]
        // calculator off
        //---------------------------------------

        if (key == KEY_OFF || key == KEY_OFF2) {
            off();
            OSTimerRestart(APD_TIMER);
        }
        else {
            return key & ~(unsigned short)0x800; // map out key repeat flag
        }
    }
}


#if 1

#define BOARD(x,y) (board[(y)*MAX_COLUMNS+(x)])

#else

//---------------------------------------------------------------
// NOTE: The following macro is very dangerous, because it
//       assumes that the compiler will not rearrange comma
//       separated commands (which is not always true)
//       [BUT I USE THIS MACRO JUST FOR A SIMPLE TEST on the VTI]
//---------------------------------------------------------------
#define BOARD(x,y)  (assert(x >= 0),assert(x < MAX_COLUMNS),assert(y >= 0),assert(y < MAX_LINES),board[(y)*MAX_COLUMNS+(x)])

#endif



/*===========================================================================*/
/* draw a line with given color from (x0,y0) to (x1,y1)                      */
/*===========================================================================*/
static void GrayDrawLine_R(short x0,short y0,short x1,short y1,short color) {
    GraySetAMSPlane(LIGHT_PLANE);
    if (color & (COLOR_WHITE | COLOR_LIGHTGRAY)) {
        DrawLine(x0, y0, x1, y1, A_NORMAL);
    }
    else {
        DrawLine(x0, y0, x1, y1, A_REVERSE);
    }
    GraySetAMSPlane(DARK_PLANE);
    if (color & (COLOR_WHITE | COLOR_DARKGRAY)) {
        DrawLine(x0, y0, x1, y1, A_NORMAL);
    }
    else {
        DrawLine(x0, y0, x1, y1, A_REVERSE);
    }
}

/*===========================================================================*/
/* fill a rectangle with given color                                         */
/*===========================================================================*/
static void FillColorRect_R(short x1,short y1,short x2, short y2, short color) {
    SCR_RECT rect;

    rect.l =   (((long)(x1))<<24)
             | (((long)(y1))<<16)
             | (((long)(x2))<<8)
             | (((long)(y2)));

    GraySetAMSPlane(LIGHT_PLANE);
    if (color & COLOR_LIGHTGRAY) {
        ScrRectFill(&rect,&fullscreen, A_NORMAL);
    }
    else {
        ScrRectFill(&rect,&fullscreen, A_REVERSE);
    }
    GraySetAMSPlane(DARK_PLANE);
    if (color & COLOR_DARKGRAY) {
        ScrRectFill(&rect,&fullscreen, A_NORMAL);
    }
    else {
        ScrRectFill(&rect,&fullscreen, A_REVERSE);
    }
}

/*===========================================================================*/
/* utility function to draw a string to both planes                          */
/* NOTE: there exists to new styles which are handled as flags               */
/*       A_CENTERED ... the given font will be used and the string will be   */
/*                      drawn centered on the screen                         */
/*       A_SHADOWED ... if this flag is set in the attributes a lightgray    */
/*                      shadow will be drawn with +1 pixel offset in all     */
/*                      two directions                                       */
/*===========================================================================*/
void GrayDrawStrExt_R(short x,short y,const char* s,short attr,short font) {
    short do_shadow = attr & A_SHADOWED;
    attr &= ~A_SHADOWED;

    short oldfont = FontGetSys();
    FontSetSys(font);

    if (attr & A_CENTERED) {
        x = (LCD_WIDTH-DrawStrWidth(s,font))/2;
        attr &= ~A_CENTERED;
    }

    GraySetAMSPlane(DARK_PLANE);
    DrawStr(x,y,s,attr);
    GraySetAMSPlane(LIGHT_PLANE);
    DrawStr(x,y,s,attr);

    if (do_shadow) DrawStr(x+1,y+1,s,A_NORMAL);
    FontSetSys(oldfont);
}



/*===========================================================================*/
/* initializes counter border                                                */
/*===========================================================================*/
static inline void InitCounterBorder(counter_t* p) {
    register short xpos=p->xpos;
    GrayDrawLine_R(xpos-1, bc_ypos-1,
                     xpos+COUNTER_WIDTH-1,bc_ypos-1,
                     COLOR_DARKGRAY);
    GrayDrawLine_R(xpos-1,bc_ypos-1,
                     xpos-1,bc_ypos+14,
                     COLOR_DARKGRAY);
    GrayDrawLine_R(xpos+COUNTER_WIDTH,bc_ypos,
                     xpos+COUNTER_WIDTH,bc_ypos+14,
                     COLOR_WHITE);
    GrayDrawLine_R(xpos, bc_ypos+15,
                     xpos+COUNTER_WIDTH,bc_ypos+15,
                     COLOR_WHITE);
}


/*===========================================================================*/
/* initializes the board graphics                                            */
/*===========================================================================*/
static inline void InitScreen(void) {
    register short board_width = ((nr_columns << 3)) + 12;
    register short b_o_x       = (((LCD_WIDTH - board_width) >> 1));
    
    //----------------------------
    // initialize global variables
    //----------------------------
    board_offset_x    = b_o_x;
    bomb_counter.xpos = b_o_x + 8;
    time_counter.xpos = b_o_x + board_width - COUNTER_WIDTH - 8;
    smiley_x          = b_o_x + (((board_width - SMILEY_WIDTH)>>1));
    tiles_offset_x    = b_o_x + 6;
    
    //----------------------------
    // clear both planes and set
    // draw_all flag to indicate
    // that each tile should be
    // drawn
    //----------------------------
    memset(GetPlane(LIGHT_PLANE),0,LCD_SIZE);
    memset(GetPlane(DARK_PLANE),0,LCD_SIZE);
    draw_all = 1;

    //----------------------------
    // draw background of board
    // and its borders
    // (DIRTY, but nevertheless
    // more efficient than
    // using bitmaps)
    //----------------------------
    
    // general background
    FillColorRect_R(b_o_x,
                  board_offset_y,
                  b_o_x+board_width-1,
                  board_offset_y+board_height,
                  COLOR_LIGHTGRAY);

    // top lines
    GrayDrawLine_R(b_o_x,board_offset_y,
                     b_o_x+board_width-2,board_offset_y,
                     COLOR_WHITE);

    GrayDrawLine_R(b_o_x,board_offset_y+1,
                     b_o_x+board_width-3,
                     board_offset_y+1,
                     COLOR_WHITE);

    GrayDrawLine_R(b_o_x+5,board_offset_y+4,
                     b_o_x+board_width-6,board_offset_y+4,
                     COLOR_DARKGRAY);

    // counter borders
    InitCounterBorder(&time_counter);
    InitCounterBorder(&bomb_counter);

    // left & right upper border
    GrayDrawLine_R(b_o_x+4,board_offset_y+4,
                     b_o_x+4,board_offset_y+23,
                     COLOR_DARKGRAY);
    GrayDrawLine_R(b_o_x+board_width-5,board_offset_y+5,
                     b_o_x+board_width-5,board_offset_y+23,
                     COLOR_WHITE);
    // horizontal lines of header
    GrayDrawLine_R(b_o_x+5,board_offset_y+24,
                     b_o_x+board_width-5,board_offset_y+24,
                     COLOR_WHITE);
    GrayDrawLine_R(b_o_x+4,board_offset_y+27,
                     b_o_x+board_width-6,board_offset_y+27,
                     COLOR_DARKGRAY);
    GrayDrawLine_R(b_o_x+4,board_offset_y+28,
                     b_o_x+board_width-7,board_offset_y+28,
                     COLOR_DARKGRAY);
    // left border lines
    GrayDrawLine_R(b_o_x,board_offset_y,
                     b_o_x,board_offset_y+board_height-1,
                     COLOR_WHITE);
    GrayDrawLine_R(b_o_x+1,board_offset_y,
                     b_o_x+1,board_offset_y+board_height-2,
                     COLOR_WHITE);
    // bottom border lines
    GrayDrawLine_R(b_o_x+2,            board_offset_y+board_height-1,
                     b_o_x+board_width-1,board_offset_y+board_height-1,
                     COLOR_DARKGRAY);
    GrayDrawLine_R(b_o_x+1,            board_offset_y+board_height,
                     b_o_x+board_width-1,board_offset_y+board_height,
                     COLOR_DARKGRAY);
    // right border lines
    GrayDrawLine_R(b_o_x+board_width-2,board_offset_y+2,
                     b_o_x+board_width-2,board_offset_y+board_height,
                     COLOR_DARKGRAY);
    GrayDrawLine_R(b_o_x+board_width-1,board_offset_y+1,
                     b_o_x+board_width-1,board_offset_y+board_height,
                     COLOR_DARKGRAY);
    // center border left
    GrayDrawLine_R(b_o_x+4,board_offset_y+28,
                     b_o_x+4,board_offset_y+board_height-5,
                     COLOR_DARKGRAY);
    GrayDrawLine_R(b_o_x+5,board_offset_y+28,
                     b_o_x+5,board_offset_y+board_height-6,
                     COLOR_DARKGRAY);
    // center border bottom
    GrayDrawLine_R(b_o_x+6,board_offset_y+board_height-5,
                     b_o_x+board_width-5,board_offset_y+board_height-5,
                     COLOR_WHITE);
    GrayDrawLine_R(b_o_x+5,board_offset_y+board_height-4,
                     b_o_x+board_width-5,board_offset_y+board_height-4,
                     COLOR_WHITE);
    // center border right
    GrayDrawLine_R(b_o_x+board_width-6,board_offset_y+29,
                     b_o_x+board_width-6,board_offset_y+board_height-5,
                     COLOR_WHITE);
    GrayDrawLine_R(b_o_x+board_width-5,board_offset_y+28,
                     b_o_x+board_width-5,board_offset_y+board_height-5,
                     COLOR_WHITE);

    //----------------------------
    // setup rest of counters and
    // output them
    //----------------------------
    bomb_counter.value = nr_bombs;
    time_counter.value = 0;
    firsttime = 1;
    InitCounter(&bomb_counter);
    InitCounter(&time_counter);
    firsttime = 0;

    //----------------------------
    // output the happy smiley ...
    //----------------------------
    OutputSmiley(SMILEY_HAPPY);
}


/*===========================================================================*/
/* push entry on stack (helper function for OpenBoardRecursive())            */
/*===========================================================================*/
void PushStack(register short x asm("%d0"),register short y asm("%d1")) {
    stack_counter++;
#if 0
    assert(stack_counter < MAX_COLUMNS*MAX_LINES);
    assert(x < MAX_COLUMNS);
    assert(y < MAX_LINES);
#endif
    stack[stack_counter].x = x;
    stack[stack_counter].y = y;
}


/*===========================================================================*/
/* pop entry from stack (helper function for OpenBoardRecursive())           */
/*===========================================================================*/
static inline stack_t* PopStack(void) {
    if (stack_counter < 0) return NULL;
    return &stack[stack_counter--];
}

/*===========================================================================*/
/* try to open board recursively (all tiles which should be opened are       */
/* marked with FLAG_TAGGED)                                                  */
/*===========================================================================*/
static void OpenBoardRecursive(void) {
    stack_t* p;

    stack_counter = -1; // initialize stack_counter

    PushStack(cursor_x,cursor_y);
    BOARD(cursor_x,cursor_y).flag = FLAG_TAGGED;

    while (p = PopStack()) {
        register short x asm("%d4");
        x = p->x;
        register short y asm("%d5");
        y = p->y;

#if 0
        assert(x >= 0);
        assert(x < nr_columns);
        assert(y >= 0);
        assert(y < nr_lines);
#endif
        if (BOARD(x,y).count <= 0) {
            //-------------------------
            // check "line above" tiles
            //-------------------------
            if (y>0) {
                if (x>0) {
                    if (BOARD(x-1,y-1).count >= 0           &&
                        BOARD(x-1,y-1).flag  == FLAG_NONE   &&
                        BOARD(x-1,y-1).state == STATE_CLOSED)
                    {
                        PushStack(x-1,y-1);
                        BOARD(x-1,y-1).flag  = FLAG_TAGGED;
                    }
                }
                if (BOARD(x,y-1).count >= 0           &&
                    BOARD(x,y-1).flag  == FLAG_NONE   &&
                    BOARD(x,y-1).state == STATE_CLOSED)
                {
                    PushStack(x,y-1);
                    BOARD(x,y-1).flag  = FLAG_TAGGED;
                }
                if (x<nr_columns-1) {
                    if (BOARD(x+1,y-1).count >= 0           &&
                        BOARD(x+1,y-1).flag  == FLAG_NONE   &&
                        BOARD(x+1,y-1).state == STATE_CLOSED)
                    {
                        PushStack(x+1,y-1);
                        BOARD(x+1,y-1).flag  = FLAG_TAGGED;
                    }
                }
            }

            //-------------------------------
            // check "within same line" tiles
            //-------------------------------
            if (x>0) {
                if (BOARD(x-1,y).count >= 0           &&
                    BOARD(x-1,y).flag  == FLAG_NONE   &&
                    BOARD(x-1,y).state == STATE_CLOSED)
                {
                    PushStack(x-1,y);
                    BOARD(x-1,y).flag  = FLAG_TAGGED;
                }
            }

            if (x<nr_columns-1) {
                if (BOARD(x+1,y).count >= 0           &&
                    BOARD(x+1,y).flag  == FLAG_NONE   &&
                    BOARD(x+1,y).state == STATE_CLOSED)
                {
                    PushStack(x+1,y);
                    BOARD(x+1,y).flag  = FLAG_TAGGED;
                }
            }

            //-------------------------------
            // check "line below" tiles
            //-------------------------------
            if (y<nr_lines-1) {
                if (x>0) {
                    if (BOARD(x-1,y+1).count >= 0           &&
                        BOARD(x-1,y+1).flag  == FLAG_NONE   &&
                        BOARD(x-1,y+1).state == STATE_CLOSED)
                    {
                        PushStack(x-1,y+1);
                        BOARD(x-1,y+1).flag  = FLAG_TAGGED;
                    }
                }
                if (BOARD(x,y+1).count >= 0           &&
                    BOARD(x,y+1).flag  == FLAG_NONE   &&
                    BOARD(x,y+1).state == STATE_CLOSED)
                {
                    PushStack(x,y+1);
                    BOARD(x,y+1).flag  = FLAG_TAGGED;
                }
                if (x<nr_columns-1) {
                    if (BOARD(x+1,y+1).count >= 0           &&
                        BOARD(x+1,y+1).flag  == FLAG_NONE   &&
                        BOARD(x+1,y+1).state == STATE_CLOSED)
                    {
                        PushStack(x+1,y+1);
                        BOARD(x+1,y+1).flag  = FLAG_TAGGED;
                    }
                }
            }
        }
    }
}


/*===========================================================================*/
/* allocates memory for board, stack and cursor background storage           */
/* NOTE: allocates as much memory as necessary for largest board, so each    */
/*       one will fit without changing the allocating procedure.             */
/*===========================================================================*/
static inline short AllocateMemory(void) {
    //------------------------------------------------------------
    // allocate memory for board
    //------------------------------------------------------------
    if (!(board = (tile_t*)malloc(sizeof(tile_t)*MAX_COLUMNS*MAX_LINES+sizeof(stack_t)*MAX_COLUMNS*MAX_LINES+2*(2*18)))) return 0;

    stack = (stack_t*)((unsigned char *)board+(sizeof(tile_t)*MAX_COLUMNS*MAX_LINES));
    cursor_buf1 = ((unsigned char *)stack+(sizeof(stack_t)*MAX_COLUMNS*MAX_LINES));
    cursor_buf2 = ((unsigned char *)cursor_buf1+2*18);

    return 1;
}


/*===========================================================================*/
/* free's allocated memory again                                             */
/*===========================================================================*/
static void FreeMemory(void) {
    if (board) {
        free(board);
//        board = NULL;
    }
}


/*===========================================================================*/
/* set all board entries to zero (initial state)                             */
/*===========================================================================*/
static inline void ClearBoard(void) {
    memset(board,0,sizeof(tile_t)*MAX_COLUMNS*MAX_LINES);
}


/*===========================================================================*/
/* distributes bombs in level (loop until all bombs are distributed)         */
/*===========================================================================*/
static inline void DistributeBombs(void) {
    short tmp = nr_bombs;
    do {
        short x = random(nr_columns);
        short y = random(nr_lines);
        if (BOARD(x,y).count == 0) {
            BOARD(x,y).count = BOMB_VALUE;
            tmp--;
        }
    }
    while (tmp > 0);
}


/*===========================================================================*/
/* calculate counts of tiles                                                 */
/*===========================================================================*/
static inline void CalculateCounts(void) {
    short x;
    short y;
    short act;

    for (y=0;y<nr_lines;y++) {
        for (x=0;x<nr_columns;x++) {
            if (BOARD(x,y).count == BOMB_VALUE) continue;

            act = 0;
            //-------------------------
            // check "line above" tiles
            //-------------------------
            if (y>0) {
                if (x>0 && BOARD(x-1,y-1).count == BOMB_VALUE)            act++;
                if (BOARD(x,y-1).count == BOMB_VALUE)                     act++;
                if (x<nr_columns-1 && BOARD(x+1,y-1).count == BOMB_VALUE) act++;
            }

            //-------------------------------
            // check "within same line" tiles
            //-------------------------------
            if (x>0 && BOARD(x-1,y).count == BOMB_VALUE) act++;
            if (x<nr_columns-1 && BOARD(x+1,y).count == BOMB_VALUE) act++;

            //-------------------------------
            // check "line below" tiles
            //-------------------------------
            if (y<nr_lines-1) {
                if (x>0 && BOARD(x-1,y+1).count == BOMB_VALUE)            act++;
                if (BOARD(x,y+1).count == BOMB_VALUE)                     act++;
                if (x<nr_columns-1 && BOARD(x+1,y+1).count == BOMB_VALUE) act++;
            }

            BOARD(x,y).count = act;
        }
    }
}


/*===========================================================================*/
/* check for finish - returns 1 if finished                                  */
/* NOTE: we could optimize this by using counters which gots decremented     */
/*       during a change, so that we haven't to scan the whole board again.  */
/*       But our calculators are that fast that we don't need such           */
/*       optimizations for a simple game like TI-Miner !!                    */
/*===========================================================================*/
static inline short IsFinished(void) {
    short x;
    short y;
    short opened = 0;
    tile_t *ptr = &BOARD(0,0);

    for (y=0;y<nr_lines;y++) {
        for (x=0;x<nr_columns;x++) {
            if (ptr->count >= 0 && ptr->state == STATE_OPENED) opened++;
            ptr++;
        }
        ptr += MAX_COLUMNS - nr_columns;
    }

    if (opened == nr_lines*nr_columns - nr_bombs) return 1;

    return 0;
}



/*===========================================================================*/
/* if tiles are tagged for update and they are closed or "un-marked" by the  */
/* user, this function will set them to state STATE_OPENED                   */
/*===========================================================================*/
static inline void OpenTaggedTiles(void) {
    short x;
    short y;
    tile_t *ptr = &BOARD(0,0);


    for (y=0;y<nr_lines;y++) {
        for (x=0;x<nr_columns;x++) {
            if (ptr->flag != FLAG_NONE) {
                if ((ptr->state == STATE_CLOSED) || (ptr->state == STATE_INTERROGATE)) {
                    ptr->state = STATE_OPENED;
                 }
            }
            ptr++;
        }
        ptr += MAX_COLUMNS - nr_columns;
    }
}



/*===========================================================================*/
/* open all bomb fields and mark them for display                            */
/*===========================================================================*/
static inline void ProcessAllBombs(void) {
    short x;
    short y;
    tile_t *ptr = &BOARD(0,0);

    for (y=0;y<nr_lines;y++) {
        for (x=0;x<nr_columns;x++) {
            if ((x == cursor_x) && (y == cursor_y)) {
                ptr->state = STATE_CLICKEDMINE;
                ptr->flag  = FLAG_TAGGED;
            }
            else if (ptr->count == BOMB_VALUE) {
                ptr->state = STATE_OPENED;
                ptr->flag  = FLAG_TAGGED;
            }
            else if (ptr->state == STATE_MARKED) {
                ptr->state = STATE_WRONGMARKED;
                ptr->flag  = FLAG_TAGGED;
            }
            ptr++;
        }
        ptr += MAX_COLUMNS - nr_columns;
    }
}



/*===========================================================================*/
/* process board drawing                                                     */
/*===========================================================================*/
static inline void DrawBoard(void) {
    short x;
    short y;
    tile_t *ptr = &BOARD(0,0);

    RestoreOldCursorPos();

    for (y=0;y<nr_lines;y++) {
        for (x=0;x<nr_columns;x++) {
            if ((ptr->flag != FLAG_NONE) || draw_all) {
                ptr->flag = FLAG_NONE;
                switch(ptr->state) {
                    case STATE_CLOSED:
                        OutputMiscTile(TYPE_UNDISCOVERED,x,y);
                        break;
                    case STATE_INTERROGATE:
                        OutputMiscTile(TYPE_INTERROGATION,x,y);
                        break;
                    case STATE_OPENED:
                        if (ptr->count >= 0) OutputCountTile(ptr->count,x,y);
                        else                 OutputMiscTile(TYPE_MINE,x,y);
                        break;
                    case STATE_MARKED:
                        OutputMiscTile(TYPE_MARKED,x,y);
                        break;
                    case STATE_WRONGMARKED:
                        OutputMiscTile(TYPE_WRONGMARK,x,y);
                        break;
                    case STATE_CLICKEDMINE:
                        OutputMiscTile(TYPE_CLICKEDMINE,x,y);
                        break;
                }
            }
            ptr++;
        }
        ptr += MAX_COLUMNS - nr_columns;
    }

    draw_all = 0;
    BackupCursorPos();
    OutputCursor();
}



/*===========================================================================*/
/* handles opening of a tile (returns 1 if a bomb tile is opened)            */
/*===========================================================================*/
static inline short HandleOpenTile(void) {
    register tile_t *ptr asm("%a3");
    ptr = &BOARD(cursor_x,cursor_y);
    switch(ptr->count) {
        case 0:
            OpenBoardRecursive();
            OpenTaggedTiles();
            return 0;
        case BOMB_VALUE:
            ptr->state = STATE_CLICKEDMINE;
            ptr->flag  = FLAG_TAGGED;
            return 1;
        default:
            ptr->state = STATE_OPENED;
            ptr->flag  = FLAG_TAGGED;
            return 0;
    }
}



/*===========================================================================*/
/* input dialog for highscore player's name                                  */
/*===========================================================================*/
static inline void GetPlayerName(void) {
    char s[PLAYER_NAME_LENGTH] = "                   ";
    short  position = 0;
    short  input;

    FillColorRect_R(8+PLANEOFFSET_X,40+PLANEOFFSET_Y,151+PLANEOFFSET_X,60+PLANEOFFSET_Y,COLOR_BLACK);
    FillColorRect_R(9+PLANEOFFSET_X,41+PLANEOFFSET_Y,150+PLANEOFFSET_X,59+PLANEOFFSET_Y,COLOR_WHITE);

    GrayDrawStrExt_R(0,43+PLANEOFFSET_Y,"New Highscore! Please enter your name:",A_REPLACE | A_CENTERED,F_4x6);

    do {
        input = GetUserInput();
        if (input == KEY_BACKSPACE) {
           if (position > 0) {
              s[--position] = 0;
              FillColorRect_R(10+PLANEOFFSET_X,52+PLANEOFFSET_Y,149+PLANEOFFSET_X,58+PLANEOFFSET_Y,COLOR_WHITE);
              GrayDrawStrExt_R(0,52+PLANEOFFSET_Y,s,A_REPLACE | A_CENTERED,F_4x6);
           }
        }
        else if (isprint(input)) {
           if (position < PLAYER_NAME_LENGTH-2) {
              s[position++] = input;
              s[position]   = 0;
              FillColorRect_R(10+PLANEOFFSET_X,52+PLANEOFFSET_Y,149+PLANEOFFSET_X,58+PLANEOFFSET_Y,COLOR_WHITE);
              GrayDrawStrExt_R(0,52+PLANEOFFSET_Y,s,A_REPLACE | A_CENTERED,F_4x6);
           }
        }
    }
    while (input != KEY_ENTER && input != KEY_ESC);

    memcpy(scores[mode].name,s,PLAYER_NAME_LENGTH);
}


/*===========================================================================*/
/* major game loop                                                           */
/*===========================================================================*/
static short PlayTheGame(/*short lines,*/short columns,short bombs) {
    short bomb_opened = 0;
    short input;

    //----------------------
    // setup game parameters
    //----------------------
    timer_started = 0;
    nr_columns    = columns;
    nr_bombs      = bombs;
    cursor_x      = 0;
    cursor_y      = 0;
    InitCursorBuffer();

    //-----------------------------------------------------------------------
    // initialize the board (clear it/distribute the bombs/calculate counts)
    //-----------------------------------------------------------------------
    ClearBoard();
    DistributeBombs();
    CalculateCounts();

    //-----------------------------------------------------
    // setup graphics related globals and initialize screen
    //-----------------------------------------------------
    InitScreen();

    //---------------------------------------
    // now loop until the user gets tired ...
    //---------------------------------------
    do {
        DrawBoard();

        input = GetUserInput();

        if (input == KEY_UP) cursor_y--;
        if (cursor_y<0) cursor_y = nr_lines-1;
        if (input == KEY_DOWN) cursor_y++;
        if (cursor_y>=nr_lines) cursor_y = 0;
        if (input == KEY_LEFT) cursor_x--;
        if (cursor_x<0) cursor_x = nr_columns-1;
        if (input == KEY_RIGHT) cursor_x++;
        if (cursor_x>=nr_columns) cursor_x = 0;

        //-----------------------------------
        // user wants to turn the active tile
        //-----------------------------------
        if (input == KEY_F1 || input == KEY_HOME) {
            tile_t *ptr = &BOARD(cursor_x,cursor_y);
/*
            if (ptr->state == STATE_INTERROGATE) {
                ptr->state = STATE_CLOSED;
            }
            if (ptr->state == STATE_CLOSED ||
                ptr->state == STATE_INTERROGATE)
            {
*/
            if (ptr->state == STATE_INTERROGATE) {
                ptr->state = STATE_CLOSED;
            }
            if (ptr->state == STATE_CLOSED)
            {
                bomb_opened = HandleOpenTile();
            }
            //---------------------------------------------------
            // setup timer to be triggered every second (20*50ms)
            // after user's first "turn tile" key press
            //---------------------------------------------------
            if (!timer_started) {
                OSRegisterTimer(USER_TIMER,20);
                timer_started = 0xFF;
            }
        }
        //-------------------------------------------
        // user wants to mark/un-mark the active tile
        //-------------------------------------------
        else if (input == KEY_ENTER || input == KEY_F2 || input == KEY_MODE) {
            tile_t *ptr = &BOARD(cursor_x,cursor_y);
            switch (ptr->state) {
                case STATE_CLOSED:
                    if (bomb_counter.value != 0) {
                        ptr->flag  = FLAG_TAGGED;
                        ptr->state = STATE_MARKED;
                        bomb_counter.value--;
                        UpdateCounter(&bomb_counter);
                    }
                    break;
                case STATE_MARKED:
                    ptr->flag  = FLAG_TAGGED;
                    ptr->state = STATE_INTERROGATE;
                    bomb_counter.value++;
                    UpdateCounter(&bomb_counter);
                    break;
                case STATE_INTERROGATE:
                    ptr->flag  = FLAG_TAGGED;
                    ptr->state = STATE_CLOSED;
                    break;
            }
            //---------------------------------------------------
            // setup timer to be triggered every second (20*50ms)
            // after user's first "mark tile" key press
            //---------------------------------------------------
            if (!timer_started) {
                OSRegisterTimer(USER_TIMER,20);
                timer_started = 0xFF;
            }
        }
        
        if (OSTimerExpired(USER_TIMER)) {
            OSTimerRestart(USER_TIMER);
            time_counter.value++;
            UpdateCounter(&time_counter);
        }
    }
    while(!IsFinished() && !bomb_opened && input != KEY_ESC);

    //------------------------------------------------------------
    // stop timer
    //------------------------------------------------------------
    if (timer_started) {
        OSFreeTimer(USER_TIMER);
        timer_started = 0;
    }

    //------------------------------------------------------------
    // if user lost the game -> show sad smiley and open all bombs
    //------------------------------------------------------------
    if (bomb_opened) {
        OutputSmiley(SMILEY_SAD);
        ProcessAllBombs();
    }
    //------------------------------------------------------------
    // if user aborts -> show the surprised smiley
    //------------------------------------------------------------
    else if (input == KEY_ESC) {
        OutputSmiley(SMILEY_SURPRISED);
    }
    else {
        OutputSmiley(SMILEY_WIN);
        //----------------------------------------------------
        // handle new highscore !!
        //----------------------------------------------------
        if (scores[mode].time >= time_counter.value) {
            DrawBoard();
            GetPlayerName();
            scores[mode].time = time_counter.value;
            WriteHighScores();
            return KEY_ENTER;
        }
    }

    //--------------------------------------------------------
    // update board ones more, remove the timer and return the
    // next keypress of the user
    //--------------------------------------------------------
    DrawBoard();

    return GetUserInput();
}


/*===========================================================================*/
/* invert a menu entry                                                       */
/*===========================================================================*/
static void InvertMenuEntry(short i) {
    SCR_RECT rect;

    /*rect.xy.x0 = 48       + PLANEOFFSET_X;
    rect.xy.y0 = 16 + 9*i + PLANEOFFSET_Y;
    rect.xy.x1 = 111      + PLANEOFFSET_X;
    rect.xy.y1 = 24 + 9*i + PLANEOFFSET_Y;*/

/*
    rect.l =   (((long)(48       + PLANEOFFSET_X))<<24)
             | (((long)(16       + PLANEOFFSET_Y))<<16)
             | (((long)(111      + PLANEOFFSET_X))<<8)
             | (((long)(24       + PLANEOFFSET_Y)));
    rect.l += 9*i + (((long)9*i)<<16);
*/

    rect.l =   (((long)(48       + PLANEOFFSET_X))<<24)
             | (((long)(16 + 9*i + PLANEOFFSET_Y))<<16)
             | (((long)(111      + PLANEOFFSET_X))<<8)
             | (((long)(24 + 9*i + PLANEOFFSET_Y)));

    GraySetAMSPlane(LIGHT_PLANE);
    ScrRectFill(&rect,&fullscreen,A_XOR);
    GraySetAMSPlane(DARK_PLANE);
    ScrRectFill(&rect,&fullscreen,A_XOR);
/*
    for (loop=0;loop<2;loop++) {
        GraySetAMSPlane(loop);
        ScrRectFill(&rect,&fullscreen,A_XOR);
    }
*/
}


/*===========================================================================*/
/* handles menu (returns 1 if game should be started)                        */
/*===========================================================================*/
short HandleMenu(void) {
    char   stmp[100];
    short  input;
    short  selection = mode;
    
    memset(GetPlane(LIGHT_PLANE),0,LCD_SIZE);
    memset(GetPlane(DARK_PLANE),0,LCD_SIZE);

    FillColorRect_R(PLANEOFFSET_X,PLANEOFFSET_Y,159+PLANEOFFSET_X,14+PLANEOFFSET_Y,COLOR_BLACK);
    GrayDrawStrExt_R(0,1+PLANEOFFSET_Y,PROGRAM_NAME" v"PROGRAM_VERSION,A_XOR | A_CENTERED,F_6x8);
    GrayDrawStrExt_R(0,9+PLANEOFFSET_Y,"\xA9 TI-Chess Team 2004",A_XOR | A_CENTERED,F_4x6);

    FillColorRect_R(PLANEOFFSET_X,62+PLANEOFFSET_Y,159+PLANEOFFSET_X,71+PLANEOFFSET_Y,COLOR_BLACK);
    GrayDrawStrExt_R(0,63+PLANEOFFSET_Y,"Highscores",A_XOR | A_CENTERED,F_6x8);

    //--------------------------------------------------------------
    // dirty hack to get the "Play 8x8" string aligned with the other
    // ones !!!
    //--------------------------------------------------------------
    //FontSetSys(F_6x8);
    GrayDrawStrExt_R(0,17+PLANEOFFSET_Y,"Play 8x8", A_REPLACE | A_CENTERED,F_6x8);
    GrayDrawStrExt_R(0,26+PLANEOFFSET_Y,"Play 8x12",A_REPLACE | A_CENTERED,F_6x8);
    GrayDrawStrExt_R(0,35+PLANEOFFSET_Y,"Play 8x16",A_REPLACE | A_CENTERED,F_6x8);
    GrayDrawStrExt_R(0,44+PLANEOFFSET_Y,"Play 8x18",A_REPLACE | A_CENTERED,F_6x8);
    GrayDrawStrExt_R(0,53+PLANEOFFSET_Y,"Exit",     A_REPLACE | A_CENTERED,F_6x8);

    //FontSetSys(F_4x6);
    sprintf(stmp,"Mode:8x8      Time:%03d    %s",scores[0].time,scores[0].name);
    GrayDrawStrExt_R(1+PLANEOFFSET_X,73+PLANEOFFSET_Y,stmp,A_REPLACE,F_4x6);
    sprintf(stmp,"Mode:8x12    Time:%03d    %s",scores[1].time,scores[1].name);
    GrayDrawStrExt_R(1+PLANEOFFSET_X,80+PLANEOFFSET_Y,stmp,A_REPLACE,F_4x6);
    sprintf(stmp,"Mode:8x16    Time:%03d    %s",scores[2].time,scores[2].name);
    GrayDrawStrExt_R(1+PLANEOFFSET_X,87+PLANEOFFSET_Y,stmp,A_REPLACE,F_4x6);
    sprintf(stmp,"Mode:8x18    Time:%03d    %s",scores[3].time,scores[3].name);
    GrayDrawStrExt_R(1+PLANEOFFSET_X,94+PLANEOFFSET_Y,stmp,A_REPLACE,F_4x6);

    InvertMenuEntry(selection);
    do {
        input = GetUserInput();
        if (input == KEY_DOWN) {
            InvertMenuEntry(selection);
            selection++;
            if (selection > 4) selection = 0;
            InvertMenuEntry(selection);
        }
        else if (input == KEY_UP) {
            InvertMenuEntry(selection);
            selection--;
            if (selection < 0) selection = 4;
            InvertMenuEntry(selection);
        }
    }
    while (input != KEY_ENTER && input != KEY_ESC);

    if (input == KEY_ENTER) {
        if (selection == 4) return 0;
        mode = selection;
        return 1;
    }

    return 0;
}


/*===========================================================================*/
/* main routine: where all the fun starts ...                                */
/*===========================================================================*/
void _main(void) {
    void *save_a5;
    asm("move.l %%a5,%0" : "=g"(save_a5));
    
    //------------------------------------------
    // turn on grayscale graphics subsystem
    //------------------------------------------
    if (!GrayOn()) return;
    
    //timer_started = 0;

    //------------------------------------------
    // try to load highscores from variable or
    // initialize with defaults
    //------------------------------------------
    LoadHighScores();

    //------------------------------------------
    // allocates memory
    //------------------------------------------
    if (!AllocateMemory()) {
        FreeMemory();
        GrayOff();
        ST_helpMsg("Error: OutOfMemory");
        GetUserInput();
/*
        ClearScreen();
        DrawStr(0,0,"Error: OutOfMemory",A_REPLACE);
        GKeyIn(NULL,0);
*/
        return;
    }

    
    //------------------------------------------
    // output logo
    //------------------------------------------
    Logo();

    //------------------------------------------
    // initialize random number generator
    //------------------------------------------
    randomize();
    
    cursor_x = cursor_y = 0;

    OSFreeTimer(USER_TIMER);
    //-------------------------------------------
    // show menu and start game depending on mode
    //-------------------------------------------
    while (HandleMenu()) {
        switch(mode) {
            case 0:  while (PlayTheGame(/*8,*/ 8,10) != KEY_ESC);  // Mode  8x8 - 10 bombs
                     break;
            case 1:  while (PlayTheGame(/*8,*/12,15) != KEY_ESC);  // Mode 12x8 - 15 bombs
                     break;
            case 2:  while (PlayTheGame(/*8,*/16,20) != KEY_ESC);  // Mode 16x8 - 20 bombs
                     break;
            default: while (PlayTheGame(/*8,*/18,25) != KEY_ESC);  // Mode 18x8 - 25 bombs
                     break;
        }
    }

    //---------------------------------------------
    // free memory and say goodbye =8-)
    //---------------------------------------------
    FreeMemory();

    GrayOff();

    asm("move.l %0,%%a5" : : "g"(save_a5));

}

//#############################################################################
//###################### NO MORE FAKES BEYOND THIS LINE #######################
//#############################################################################
//
//=============================================================================
// Revision History
//=============================================================================
//
// $Log: timiner.c,v $
// Revision 1.3  2004/07/14 15:08:58  DEBROUX Lionel
// recompiled with TIGCC 0.95 Beta 12
//
// Revision 1.2  2004/07/11 16:45:03  DEBROUX Lionel
// generic commit
//
// Revision 1.1  2002/09/27 10:10:30  tnussb
// initial check-in to CVS repository
//
//
